@@ -174,23 +174,40 @@ matattr <- function(free, est, constraint, mat, Ng, std.lv, wig, ...) {
   return(out)
 }
 
-lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=NULL, prisamp=FALSE, mcmcextra=NULL) {
+lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=NULL, prisamp=FALSE, mcmcextra=NULL, level=1L, indat=NULL) {
   ## extract model and data characteristics from lavaan object
   opts <- lavInspect(lavobject, 'options')
+  multilevel <- opts$.multilevel
+
+  ## if not multilevel, this creates empty matrices to pass to stan for level 2
+  ## if (!multilevel & level > 1) stop("blavaan ERROR: higher levels requested, but this is not a multilevel model.")
 
   ## data characteristics
-  dat <- lav2standata(lavobject)
-  Ng <- dat$Ng
-  
-  ## model
-  if ("emiter" %in% names(mcmcextra$data)) {
-    dat$emiter <- mcmcextra$data$emiter
+  if (length(indat) == 0) {
+    dat <- lav2standata(lavobject)
+    ## model
+    if ("emiter" %in% names(mcmcextra$data)) {
+      dat$emiter <- mcmcextra$data$emiter
+    } else {
+      dat$emiter <- 20L
+    }
+    dat$pri_only <- prisamp
+    Ng <- dat$Ng
   } else {
-    dat$emiter <- 20L
+    dat <- list()
+    Ng <- indat$Ng
   }
-  dat$pri_only <- prisamp
+
   freemats <- lavInspect(lavobject, 'free')
   constrain <- attr(freemats, 'header')
+  if (multilevel) {
+    freemats <- freemats[2*(1:Ng) - 2 + level]
+  } else if (level == 2L) {
+    freemats <- list(0)
+  } else if (inherits(freemats[[1]], "matrix")) {
+    freemats <- list(freemats)
+  }
+
   if (any(constrain$op != "==")) {
     ops <- unique(constrain$op)
     ops <- ops[ops != "=="]
@@ -198,15 +215,30 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
                paste(ops, collapse=" "), "\n Try target='stanclassic'"))
   }
   estmats <- lavInspect(lavobject, 'est')
-  if (Ng == 1) {
-    freemats <- list(freemats)
+  if (multilevel) {
+    estmats <- estmats[2*(1:Ng) - 2 + level]
+  } else if (level == 2L) {
+    estmats <- list(0)
+  } else if (inherits(estmats[[1]], "matrix")) {
     estmats <- list(estmats)
   }
+
   free2 <- list()
   nfree <- list()
   lavpartable <- parTable(lavobject)
   lavpartable <- lavMatrixRepresentation(lavpartable, add.attributes = TRUE)
-  lavpartable <- lavpartable[order(lavpartable$group, lavpartable$col, lavpartable$row),]
+  if (multilevel & level == 1L) {
+    lavpartable <- subset(lavpartable, level == "within")
+  } else if (multilevel & level == 2L) {
+    lavpartable <- subset(lavpartable, level == "between")
+  } else if (level == 2L) {
+    lavpartable <- lavpartable[0,]
+  }
+  if ("group" %in% names(lavpartable)) {
+    lavpartable <- lavpartable[order(lavpartable$group, lavpartable$col, lavpartable$row),]
+  } else {
+    lavpartable <- lavpartable[order(lavpartable$col, lavpartable$row),]
+  }
 
   if (length(wiggle) > 0){
     wigls <- wiglabels(lavpartable, wiggle, wiggle.sd)
@@ -241,7 +273,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(lambda = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'lambda'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
@@ -252,36 +285,38 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
 
   ## 2. Lambda_x; never used because x only pops up in
   ##    the conditional case.
-  dat$Lambda_x_skeleton <- array(0, dim = c(Ng, 0, 0))
-  dat$w2skel <- matrix(0, 0, 3)
-  dat$lam_x_sign <- matrix(0, 0, 2)
+  if (level == 1L) {
+    dat$Lambda_x_skeleton <- array(0, dim = c(Ng, 0, 0))
+    dat$w2skel <- matrix(0, 0, 3)
+    dat$lam_x_sign <- matrix(0, 0, 2)
 
-  ## 3. Gamma
-  if ("gamma" %in% names(freemats[[1]])) {
-    fr <- lapply(freemats, function(x) x$gamma)
-    es <- lapply(estmats, function(x) x$gamma)
-    frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
-    twsel <- lavpartable$free %in% frnums
-    tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
-    ## Note: if Lambda_x were in use, then FALSE below needs to
-    ##       be opts$std.lv:
-    res <- matattr(fr, es, constrain, mat = "Gamma", Ng, FALSE, tmpwig)
+    ## 3. Gamma
+    if ("gamma" %in% names(freemats[[1]])) {
+      fr <- lapply(freemats, function(x) x$gamma)
+      es <- lapply(estmats, function(x) x$gamma)
+      frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
+      twsel <- lavpartable$free %in% frnums
+      tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
+      ## Note: if Lambda_x were in use, then FALSE below needs to
+      ##       be opts$std.lv:
+      res <- matattr(fr, es, constrain, mat = "Gamma", Ng, FALSE, tmpwig)
 
-    dat$Gamma_skeleton <- res$matskel
-    dat$w3skel <- res$wskel
-    dat$gam_sign <- res$sign
-    free2 <- c(free2, list(gamma = res$free))
-    ptrows <- which(lavpartable$mat == "gamma" & lavpartable$free > 0)
-    veclen <- length(ptrows)
-    if (veclen > 0) {
-      fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(gamma = sum(fpars)))
-      freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      dat$Gamma_skeleton <- res$matskel
+      dat$w3skel <- res$wskel
+      dat$gam_sign <- res$sign
+      free2 <- c(free2, list(gamma = res$free))
+      ptrows <- which(lavpartable$mat == "gamma" & lavpartable$free > 0)
+      veclen <- length(ptrows)
+      if (veclen > 0) {
+        fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
+        nfree <- c(nfree, list(gamma = sum(fpars)))
+        freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      }
+    } else {
+      dat$Gamma_skeleton <- array(0, dim = c(Ng, dim(dat$Lambda_y_skeleton)[3], 0))
+      dat$w3skel <- matrix(0, 0, 3)
+      dat$gam_sign <- matrix(0, 0, 3)
     }
-  } else {
-    dat$Gamma_skeleton <- array(0, dim = c(Ng, dim(dat$Lambda_y_skeleton)[3], 0))
-    dat$w3skel <- matrix(0, 0, 3)
-    dat$gam_sign <- matrix(0, 0, 3)
   }
 
   ## 4. Beta
@@ -302,7 +337,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(beta = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'beta'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
@@ -339,7 +375,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(theta = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'theta' 
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
@@ -375,85 +412,88 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(rho = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'rho'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
     dat$Theta_r_skeleton <- array(0, dim = c(Ng, 0, 0))
     dat$w7skel <- matrix(0, 0, 3)
   }
-  
-  ## 6. diag(Theta_x)
-  if ("cov.x" %in% names(freemats[[1]])) {
-    fr <- lapply(freemats, function(x){
-      dmat <- x$cov.x
-      dmat[lower.tri(dmat)] <- dmat[upper.tri(dmat)] <- 0
-      dmat}
-      )
-    
-    es <- lapply(estmats, function(x){
-      dmat <- x$cov.x
-      dmat[lower.tri(dmat)] <- dmat[upper.tri(dmat)] <- 0
-      dmat}
-      )
-    dest <- es
 
-    frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
-    twsel <- lavpartable$free %in% frnums
-    tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
+  if (level == 1L) {
+    ## 6. diag(Theta_x)
+    if ("cov.x" %in% names(freemats[[1]])) {
+      fr <- lapply(freemats, function(x){
+        dmat <- x$cov.x
+        dmat[lower.tri(dmat)] <- dmat[upper.tri(dmat)] <- 0
+        dmat}
+        )
     
-    res <- matattr(fr, es, constrain, mat = "Theta_x", Ng, opts$std.lv, tmpwig)
+      es <- lapply(estmats, function(x){
+        dmat <- x$cov.x
+        dmat[lower.tri(dmat)] <- dmat[upper.tri(dmat)] <- 0
+        dmat}
+        )
+      dest <- es
 
-    dat$Theta_x_skeleton <- res$matskel
-    dat$w6skel <- res$wskel
-    free2 <- c(free2, list(cov.x = res$free))
-    ptrows <- with(lavpartable, which(mat == "cov.x" & free > 0 & row == col))
-    veclen <- length(ptrows)
-    if (veclen > 0) {
-      fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(cov.x = sum(fpars)))
-      freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
+      twsel <- lavpartable$free %in% frnums
+      tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
+    
+      res <- matattr(fr, es, constrain, mat = "Theta_x", Ng, opts$std.lv, tmpwig)
+
+      dat$Theta_x_skeleton <- res$matskel
+      dat$w6skel <- res$wskel
+      free2 <- c(free2, list(cov.x = res$free))
+      ptrows <- with(lavpartable, which(mat == "cov.x" & free > 0 & row == col))
+      veclen <- length(ptrows)
+      if (veclen > 0) {
+        fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
+        nfree <- c(nfree, list(cov.x = sum(fpars)))
+        freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      }
+    } else {
+      dat$Theta_x_skeleton <- array(0, dim = c(Ng, 0, 0))
+      dat$w6skel <- matrix(0, 0, 3)
     }
-  } else {
-    dat$Theta_x_skeleton <- array(0, dim = c(Ng, 0, 0))
-    dat$w6skel <- matrix(0, 0, 3)
-  }
 
 
-  ## 8. Theta_x_r
-  if ("cov.x" %in% names(freemats[[1]])) {
-    fr <- lapply(freemats, function(x){
-      dmat <- x$cov.x
-      diag(dmat) <- 0L
-      dmat}
-      )
+    ## 8. Theta_x_r
+    if ("cov.x" %in% names(freemats[[1]])) {
+      fr <- lapply(freemats, function(x){
+        dmat <- x$cov.x
+        diag(dmat) <- 0L
+        dmat}
+        )
     
-    es <- lapply(estmats, function(x){
-      dmat <- x$cov.x
-      diag(dmat) <- 1L
-      dmat[upper.tri(dmat)] <- 0L
-      dmat}
-      )
+      es <- lapply(estmats, function(x){
+        dmat <- x$cov.x
+        diag(dmat) <- 1L
+        dmat[upper.tri(dmat)] <- 0L
+        dmat}
+        )
 
-    frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
-    twsel <- lavpartable$free %in% frnums
-    tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
-    
-    res <- matattr(fr, es, constrain, mat = "Theta_x_r", Ng, opts$std.lv, tmpwig, dest = dest)
+      frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
+      twsel <- lavpartable$free %in% frnums
+      tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
+      
+      res <- matattr(fr, es, constrain, mat = "Theta_x_r", Ng, opts$std.lv, tmpwig, dest = dest)
 
-    dat$Theta_x_r_skeleton <- res$matskel
-    dat$w8skel <- res$wskel
-    free2 <- c(free2, list(cov.x = res$free))
-    ptrows <- with(lavpartable, which(mat == "cov.x" & free > 0 & row != col))
-    veclen <- length(ptrows)
-    if (veclen > 0) {
-      fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(cov.x = sum(fpars)))
-      freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      dat$Theta_x_r_skeleton <- res$matskel
+      dat$w8skel <- res$wskel
+      free2 <- c(free2, list(cov.x = res$free))
+      ptrows <- with(lavpartable, which(mat == "cov.x" & free > 0 & row != col))
+      veclen <- length(ptrows)
+      if (veclen > 0) {
+        fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
+        nfree <- c(nfree, list(cov.x = sum(fpars)))
+        freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      }
+    } else {
+      dat$Theta_x_r_skeleton <- array(0, dim = c(Ng, 0, 0))
+      dat$w8skel <- matrix(0, 0, 3)
     }
-  } else {
-    dat$Theta_x_r_skeleton <- array(0, dim = c(Ng, 0, 0))
-    dat$w8skel <- matrix(0, 0, 3)
   }
 
   ## 9. diag(Psi)
@@ -485,7 +525,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if(veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(psi = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'psi'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
@@ -494,6 +535,7 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
   }
 
   ## 10. Psi_r
+  dat$fullpsi <- 0L
   if ("psi" %in% names(freemats[[1]])) {
     fr <- lapply(freemats, function(x){
       dmat <- x$psi
@@ -516,7 +558,6 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
                    free1 = free2$lambda, free2 = lyfree2, sign = dat$lam_y_sign,
                    dest = dest)
 
-    dat$fullpsi <- 0L
     dat$Psi_r_skeleton <- res$matskel
     dat$w10skel <- res$wskel
     dat$psi_r_sign <- res$sign
@@ -525,7 +566,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(lvrho = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'lvrho'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
     ## check for completely unrestricted correlation matrix, for lkj
@@ -540,14 +582,16 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     dat$psi_r_sign <- matrix(0, 0, 3)
   }
 
-  ## 11. Phi unused
-  dat$Phi_skeleton <- array(0, dim = c(Ng, 0, 0))
-  dat$w11skel <- matrix(0, 0, 3)
+  if (level == 1L) {
+    ## 11. Phi unused
+    dat$Phi_skeleton <- array(0, dim = c(Ng, 0, 0))
+    dat$w11skel <- matrix(0, 0, 3)
 
-  ## 12. Phi_r unused
-  dat$Phi_r_skeleton <- array(0, dim = c(Ng, 0, 0))
-  dat$w12skel <- matrix(0, 0, 3)
-  dat$phi_r_sign <- matrix(0, 0, 3)
+    ## 12. Phi_r unused
+    dat$Phi_r_skeleton <- array(0, dim = c(Ng, 0, 0))
+    dat$w12skel <- matrix(0, 0, 3)
+    dat$phi_r_sign <- matrix(0, 0, 3)
+  }
 
   ## 13. Nu NB: unlike lavaan, we paste mean.x to end!!
   if ("nu" %in% names(freemats[[1]])) {
@@ -575,7 +619,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(nu = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'nu'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)      
     }
   } else {
@@ -601,7 +646,8 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     veclen <- length(ptrows)
     if (veclen > 0) {
       fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(alpha = sum(fpars)))
+      nfree <- c(nfree, list(sum(fpars)))
+      names(nfree)[length(nfree)] <- 'alpha'
       freeparnums[ptrows[fpars]] <- 1:sum(fpars)
     }
   } else {
@@ -610,30 +656,32 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
   }
 
   ## 15. Tau
-  if ("tau" %in% names(freemats[[1]])) {
-    fr <- lapply(freemats, function(x) x$tau)
-    es <- lapply(estmats, function(x) x$tau)
+  if (level == 1L) {
+    if ("tau" %in% names(freemats[[1]])) {
+      fr <- lapply(freemats, function(x) x$tau)
+      es <- lapply(estmats, function(x) x$tau)
 
-    frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
-    twsel <- lavpartable$free %in% frnums
-    tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
+      frnums <- sapply(fr, function(x) as.numeric(x[x != 0]))
+      twsel <- lavpartable$free %in% frnums
+      tmpwig <- lavpartable[twsel,'free'][which(lavpartable[twsel, 'plabel'] %in% wig)]
 
-    res <- matattr(fr, es, constrain, mat = "Tau", Ng, opts$std.lv, tmpwig)
-    dat$Tau_skeleton <- res$matskel
-    dat$w15skel <- res$wskel
-    free2 <- c(free2, list(tau = res$free))
-    ptrows <- with(lavpartable, which(mat == "tau" & free > 0))
-    veclen <- length(ptrows)
-    if (veclen > 0) {
-      fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
-      nfree <- c(nfree, list(tau = sum(fpars)))
-      freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      res <- matattr(fr, es, constrain, mat = "Tau", Ng, opts$std.lv, tmpwig)
+      dat$Tau_skeleton <- res$matskel
+      dat$w15skel <- res$wskel
+      free2 <- c(free2, list(tau = res$free))
+      ptrows <- with(lavpartable, which(mat == "tau" & free > 0))
+      veclen <- length(ptrows)
+      if (veclen > 0) {
+        fpars <- res$wskel[1:veclen,1] == 0 | res$wskel[1:veclen,3] == 1
+        nfree <- c(nfree, list(tau = sum(fpars)))
+        freeparnums[ptrows[fpars]] <- 1:sum(fpars)
+      }
+    } else {
+      dat$Tau_skeleton <- array(0, dim = c(Ng, 0, 0))
+      dat$w15skel <- matrix(0, 0, 3)
     }
-  } else {
-    dat$Tau_skeleton <- array(0, dim = c(Ng, 0, 0))
-    dat$w15skel <- matrix(0, 0, 3)
   }
-  
+
   ## add priors by using set_stanpars() from classic approach
   ## needs some partable mods
   lavpartable$rhoidx <- rep(0, length(lavpartable$mat))
@@ -656,8 +704,11 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
   lpt <- lavpartable
   lpt$mat[lpt$op == ":="] <- "def"
   dp <- c(dp, def = "")
-  stanprires <- set_stanpars("", lpt, prifree, dp, ov.names)
-  lavpartable$prior <- stanprires$partable$prior
+
+  if (nrow(lpt) > 0) {
+    stanprires <- set_stanpars("", lpt, prifree, dp, "")
+    lavpartable$prior <- stanprires$partable$prior
+  }
 
   dat$wigind <- 0L
   if (length(wig) > 0) {
@@ -671,7 +722,7 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
   lavpartable$freeparnums <- freeparnums
 
   ## FIXME theta_x, cov.x not handled
-  if (!(inits %in% c("jags", "stan"))) {
+  if (!(inits %in% c("jags", "stan")) & nrow(lavpartable) > 0) {
     ini <- set_inits_stan(lavpartable, nfree, n.chains, inits)
 
     mapping <- c(Lambda_y_free = "lambdafree", Gamma_free = "gammafree",
@@ -683,27 +734,37 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
     for (i in 1:length(ini)) {
       nmidx <- match(names(ini[[i]]), mapping)
       names(ini[[i]]) <- names(mapping)[nmidx]
-      if(dat$fullpsi) {
+      if (dat$fullpsi) {
         ## remove Psi_r_free because handled as corr_mat
         ini[[i]]$Psi_r_free <- array(0, 0)
         psidim <- dim(dat$Psi_skeleton)[2]
-        psimat <- array(diag(1, psidim), dim = c(psidim, psidim, dat$Ng))
+        psimat <- array(diag(1, psidim), dim = c(psidim, psidim, Ng))
         ini[[i]]$Psi_r_mat <- aperm(psimat, perm = c(3, 1, 2))
       }
-      ## tau needs a specific ordering, with augmented z's to match
-      tauvec <- which(names(ini[[i]]) == "Tau_free")
-      if(length(tauvec) > 0) {
-        z_aug <- rep(.5, dat$Noent)
 
-        ## for (j in 1:dat$Nord) {
-        ##   tmpyx <- dat$YXo[,j]
-        ##   hicat <- tmpyx == max(tmpyx)
-        ##   locat <- tmpyx == min(tmpyx)
-        ##   z_aug[hicat,j] <- .05
-        ##   z_aug[locat,j] <- .95
-        ## }
+      ## for target = "stan", translate correlation parameters from (0,1) to (-1,1)
+      ## (needed because set_inits_stan() is also used for stanclassic and stancond targets
+      if ("Theta_r_free" %in% names(ini[[i]])) ini[[i]]$Theta_r_free <- -1 + 2 * ini[[i]]$Theta_r_free
+      if ("Psi_r_free" %in% names(ini[[i]])) ini[[i]]$Psi_r_free <- -1 + 2 * ini[[i]]$Psi_r_free
+        
+      if (level == 2L) {
+        names(ini[[i]]) <- paste0(names(ini[[i]]), "_c")
+      } else {
+        ## if ordinal, tau needs a specific ordering, with augmented z's to match
+        tauvec <- which(names(ini[[i]]) == "Tau_free")
+        if(length(tauvec) > 0) {
+          z_aug <- rep(.5, dat$Noent)
 
-        ini[[i]] <- c(ini[[i]], list(z_aug = z_aug))
+          ## for (j in 1:dat$Nord) {
+          ##   tmpyx <- dat$YXo[,j]
+          ##   hicat <- tmpyx == max(tmpyx)
+          ##   locat <- tmpyx == min(tmpyx)
+          ##   z_aug[hicat,j] <- .05
+          ##   z_aug[locat,j] <- .95
+          ## }
+
+          ini[[i]] <- c(ini[[i]], list(z_aug = z_aug))
+        }
       }
     }
   } else {
@@ -711,19 +772,33 @@ lav2stanmarg <- function(lavobject, dp, n.chains, inits, wiggle=NULL, wiggle.sd=
   }
 
   ## index of dummy lvs, for sampling lvs
-  dumlv <- c(lavobject@Model@ov.x.dummy.lv.idx[[1]],
-             lavobject@Model@ov.y.dummy.lv.idx[[1]])
+  dumlv <- NULL
+  if (level == 1L) {
+    dumlv <- c(lavobject@Model@ov.x.dummy.lv.idx[[1]],
+               lavobject@Model@ov.y.dummy.lv.idx[[1]])
+  } else if (multilevel) {
+    dumlv <- c(lavobject@Model@ov.x.dummy.lv.idx[[2]],
+               lavobject@Model@ov.y.dummy.lv.idx[[2]])
+  }
+
+  ## for level 2, add _c to names
+  if (level == 2L) {
+    names(dat) <- paste0(names(dat), "_c")
+    if (length(free2) > 0) names(free2) <- paste0(names(free2), "_c")
+  }
   
   return(list(dat = dat, free2 = free2, lavpartable = lavpartable,
               init = ini, dumlv = dumlv, wigpris = wigpris))
 }
 
 
-coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "mean") {
+coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, dmnames, level = 1L, fun = "mean") {
   ## Extract posterior means from marginal stan model.
   ## free2 comes from lav2lers().
   ## lersdat is data passed to sem stan code.
   ## rsob is the result of sampling().
+  if (!(level %in% c(1L, 2L))) stop("blavaan ERROR: Bad level specification in coeffun().", call. = FALSE)
+  
   stanfit <- !is.null(rsob)
   if(stanfit){
     rssumm <- rstan::summary(rsob)
@@ -734,10 +809,12 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
       b.est <- rssumm$summary[,"mean"]
     } else if(fun == "median"){
       b.est <- rssumm$summary[,"50%"]
+    } else {
+      stop(paste0("blavaan ERROR: ", fun, " not implemented in coeffun()."), call. = FALSE)
     }
     sd.est <- rssumm$summary[,"sd"]
   }
-  
+
   ## lavaan pars to stan par vectors
   mapping <- c(ly_sign = "lambda", g_sign = "gamma",
                bet_sign = "beta", Theta_cov = "theta",
@@ -745,6 +822,15 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
                Theta_x_var = "cov.x", Psi_cov = "psi",
                Psi_var = "psi", Nu_free = "nu", ## includes mean.x!
                Alpha_free = "alpha", Tau_free = "tau")
+  matmod <- ""
+  olpt <- lavpartable
+  if (level == 2L) {
+    names(mapping) <- paste0(names(mapping), "_c")
+    lavpartable <- lapply(lavpartable, function(x) x[lavpartable$level == "between"])
+    matmod <- "_c"
+  } else if ("level" %in% names(lavpartable)) {
+    lavpartable <- lapply(lavpartable, function(x) x[lavpartable$level == "within"])
+  }
 
   ## lavaan pars to w?skel (for equality constraints)
   mapping2 <- c("lambda", "gamma", "beta", "theta",
@@ -760,11 +846,15 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
                 dtheta_x = "Theta_x_var", rpsi = "Psi_cov",
                 dpsi = "Psi_var", nu = "Nu_free",
                 alpha = "Alpha_free", tau = "Tau_free")
+  if (level == 2L) {
+    mapping3 <- sapply(mapping3, function(x) paste0(x, "_c"))
+    names(mapping3) <- sapply(names(mapping3), function(x) paste0(x, "_c"))
+  }
 
   ## check names in lavfree
   deltloc <- which(names(lavfree) == "delta")
   if(length(deltloc) > 0) lavfree <- lavfree[-deltloc]
-  if(!all(names(lavfree) %in% mapping)){
+  if(!all(names(lavfree) %in% mapping) || is.null(names(lavfree))){
     ## multiple groups? FIXME handle delta
     deltloc <- which(names(lavfree[[1]]) == "delta")
     if(length(deltloc) > 0) lavfree <- lapply(lavfree, function(x) x[-deltloc])
@@ -780,13 +870,14 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
 
   freeidx <- lapply(lavfree, function(x) lapply(x, function(y) which(y > 0, arr.ind = TRUE)))
   freenums <- lapply(free2, function(x) lapply(x, function(y) y[y > 0]))
-  nfree <- max(sapply(lavfree, function(x)
-    sapply(x, function(x) ifelse(length(x) > 0, max(x), 0))))
+  tmpfree <- unlist(lavfree)
+  nfree <- length(unique(tmpfree[tmpfree > 0]))
+  if(any(lavpartable$free > 0)){
+    freevec <- lavpartable$free[lavpartable$free > 0]
+  }
 
+  vcorr <- NULL
   if(stanfit){
-    draw_mat <- as.matrix(rsob)
-
-    freevec <- rep(NA, nfree)
     rowidx <- rowidx2 <- rep(NA, nfree) # row index of stan est and summary containing the parameters (for vcorr)
 
     ## 1. get free par vector
@@ -794,20 +885,20 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
     ## 3. fill "x" in using lavaan free
     ## 4. record freeidx, double-counting free parameters
     est <- sdvec <- rep(NA, nfree)
-
     for(m in 1:length(freeidx[[1]])){
       stanvec <- names(mapping)[mapping == names(freeidx[[1]])[m]]
       wskel <- names(mapping2)[mapping == names(freeidx[[1]])[m]]
-      wvec <- paste0("w", wskel)
-      wgvec <- paste0("wg", wskel)
-      wskel <- paste0(wvec, "skel")
+      wvec <- paste0("w", wskel, matmod)
+      wgvec <- paste0("wg", wskel, matmod)
+      wskel <- paste0("w", wskel, "skel", matmod)
 
       ## 2 for cov/var vectors, 1 otherwise
       if(length(stanvec) > 2) stop("blavaan ERROR: problem with mapping from stan to lavaan")
 
       for(j in 1:length(stanvec)){
         freename <- names(mapping3)[mapping3 == stanvec[j]]
-        parnums <- do.call("c", freenums[[freename]])      
+        parnums <- do.call("c", freenums[[freename]])
+        parnums <- match(parnums, freevec)
         tmpw <- lersdat[[wskel[j]]]
 
         if(is.na(parnums[1])) next
@@ -823,8 +914,13 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
           ## ordered differently from stan draws rows
           samppar <- (tmpw[,1] == 0) | (tmpw[,3] == 1) # free or constrained prior
           parvec <- tmpsd <- rowvec <- rowvec2 <- rep(NA, NROW(tmpw))
-          rowvec[samppar] <- grep(stanvec[j], names(b.est))
-          rowvec2[samppar] <- grep(stanvec[j], colnames(draw_mat))
+          if(level == 1L){
+            rowvec[samppar] <- which(grepl(stanvec[j], names(b.est)) & !(grepl("_c\\[", names(b.est))))
+            rowvec2[samppar] <- which(grepl(stanvec[j], dmnames) & !(grepl("_c\\[", dmnames)))
+          } else {
+            rowvec[samppar] <- grep(stanvec[j], names(b.est))
+            rowvec2[samppar] <- grep(stanvec[j], dmnames)
+          }
           parvec[samppar] <- b.est[rowvec[samppar]]
           tmpsd[samppar] <- sd.est[rowvec[samppar]]
 
@@ -843,9 +939,7 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
       }
     }
 
-    vcorr <- cor(draw_mat[, rowidx2, drop=FALSE])
-
-    names(sdvec) <- colnames(vcorr)
+    names(sdvec) <- dmnames[rowidx2]
 
     ## add to partable for other functions
     ## indexing of stan objects
@@ -864,12 +958,22 @@ coeffun_stanmarg <- function(lavpartable, lavfree, free2, lersdat, rsob, fun = "
     lavpartable$pxnames[lavpartable$free > 0] <- rownames(rssumm$summary)[rowidx2]
   } else {
     sdvec <- NULL
-    vcorr <- NULL
     rssumm <- list(summary = NULL)
   }
   
   ## matrices and names
-  lavpartable <- lavMatrixRepresentation(lavpartable, add.attributes = TRUE, as.data.frame. = FALSE)
+  if("level" %in% names(lavpartable)){
+    olpt <- lavMatrixRepresentation(olpt, add.attributes = TRUE, as.data.frame. = FALSE)
+
+    if(level == 2L){
+      olpt <- lapply(olpt, function(x) x[olpt$level == "between"])
+    } else {
+      olpt <- lapply(olpt, function(x) x[olpt$level == "within"])
+    }
+    lavpartable <- c(lavpartable, list(mat = olpt$mat, row = olpt$row, col = olpt$col))
+  } else {
+    lavpartable <- lavMatrixRepresentation(lavpartable, add.attributes = TRUE, as.data.frame. = FALSE)
+  } 
   
   list(x = lavpartable$est[lavpartable$free > 0],
        lavpartable = lavpartable,
@@ -891,11 +995,27 @@ lav2standata <- function(lavobject) {
   } else {
     nvar <- ncol(YX[[1]])
   }
+
   ord <- as.numeric(lavInspect(lavobject, 'categorical'))
+  multilevel <- lavInspect(lavobject, 'options')$.multilevel
+  if (multilevel) Lp <- lavobject@Data@Lp
   dat$ord <- ord
+  dat$multilev <- multilevel
   dat$N <- lavInspect(lavobject, 'nobs')
 
-  xidx <- lavobject@SampleStats@x.idx[[1]]
+  if (multilevel) {
+    dat$ov_idx1 <- Lp[[1]]$ov.idx[[1]]
+    dat$ov_idx2 <- Lp[[1]]$ov.idx[[2]]
+    dat$p_tilde <- length(unique(c(dat$ov_idx1, dat$ov_idx2)))
+
+    xidx <- Lp[[1]]$ov.x.idx[[1]]
+    if (is.null(xidx)) xidx <- integer(0)
+    xidxb <- Lp[[1]]$ov.x.idx[[2]]
+    if (is.null(xidxb)) xidxb <- integer(0)
+  } else {
+    xidx <- lavobject@SampleStats@x.idx[[1]]
+    xidxb <- integer(0)
+  }
   allvars <- 1:nvar
 
   ## lavobject@SampleStats@missing.flag is TRUE when missing='ml',
@@ -924,6 +1044,10 @@ lav2standata <- function(lavobject) {
     dat$Nx <- rep(0, dat$Np)
     dat$Xvar <- dat$Xdatvar <- matrix(0, dat$Np, nvar)
 
+    ## these are placeholders because we do not yet allow missingness in two-level models:
+    dat$Nx_between <- rep(0, dat$Np)
+    dat$Xbetvar <- matrix(0, dat$Np, nvar)
+
     for (i in 1:dat$Np) {
       dat$Obsvar[i, 1:dat$Nobs[i]] <- Obsvar[[i]]
       if (dat$Nobs[i] < nvar) {
@@ -959,18 +1083,144 @@ lav2standata <- function(lavobject) {
     dat$Np <- dat$Ng
     dat$Nobs <- array(nvar, dat$Np)
     dat$Obsvar <- matrix(1:nvar, dat$Np, nvar, byrow=TRUE)
+    if (multilevel) {
+      ptot <- length(unique(c(Lp[[1]]$ov.idx[[1]]))) #, Lp$ov.idx[[2]])))
+      dat$Obsvar <- matrix(1:ptot, dat$Np, ptot, byrow=TRUE)
+      dat$Nobs <- array(ptot, dat$Np)
+    }
     dat$Nx <- array(length(xidx), dat$Np)
-    dat$Xvar <- dat$Xdatvar <- matrix(xidx, dat$Np, length(xidx), byrow=TRUE)
-    if (length(xidx) < nvar) {
-      dat$Xvar <- dat$Xdatvar <- cbind(dat$Xvar,
-                                       matrix(allvars[!(allvars %in% xidx)], dat$Np,
-                                              nvar - length(xidx), byrow = TRUE))
+    dat$Nx_between <- array(length(xidxb), dat$Np)
+
+    dat$Xvar <- dat$Xdatvar <- matrix(xidx, dat$Np, length(xidx), byrow = TRUE)
+    dat$Xbetvar <- matrix(xidxb, dat$Np, length(xidxb), byrow = TRUE)
+    if (multilevel) {
+      if (length(xidx) < length(allvars)) {
+        dat$Xvar <- dat$Xdatvar <- cbind(dat$Xvar,
+                                         matrix(allvars[!(allvars %in% xidx)], dat$Np,
+                                                length(allvars) - length(xidx), byrow = TRUE))
+      }
+      if (length(xidxb) < length(allvars)) {
+        dat$Xbetvar <- cbind(dat$Xbetvar,
+                             matrix(allvars[!(allvars %in% xidxb)], dat$Np,
+                                    length(allvars) - length(xidxb), byrow = TRUE))
+      }
+    } else {
+      if (length(xidx) < nvar) {
+        dat$Xvar <- dat$Xdatvar <- cbind(dat$Xvar,
+                                         matrix(allvars[!(allvars %in% xidx)], dat$Np,
+                                                nvar - length(xidx), byrow = TRUE))
+      }
+      dat$Xbetvar <- matrix(0, dat$Np, nvar, byrow = TRUE)
     }
   }
   dat$YX <- do.call("rbind", YX)
   dat$S <- S
   dat$grpnum <- array(dat$grpnum, length(dat$grpnum))
 
+  if (multilevel) {
+    ## NB: Lp has one list entry per group
+    dat$nclus <- array(t(sapply(Lp, function(x) unlist(x$nclusters))), dim = c(Ng, 2))
+    ## these are in one vector to avoid ragged arrays in Stan. We need ncluster_sizes to decide
+    ## what sizes belong to which group.
+    dat$cluster_size <- unlist(sapply(Lp, function(x) x$cluster.size[[2]], simplify = FALSE)) 
+    dat$cluster_sizes <- unlist(sapply(Lp, function(x) x$cluster.sizes[[2]], simplify = FALSE))
+    dat$ncluster_sizes <- array(sapply(Lp, function(x) length(x$cluster.sizes[[2]])), Ng)
+    dat$cluster_size_ns <- unlist(sapply(Lp, function(x) x$cluster.size.ns[[2]], simplify = FALSE))
+    ## we assume variable indices are the same across groups, this could be revisited later
+    dat$between_idx <- array(Lp[[1]]$between.idx[[2]], length(Lp[[1]]$between.idx[[2]]))
+    dat$N_between <- length(dat$between_idx)
+    dat$within_idx <- array(Lp[[1]]$within.idx[[2]], length(Lp[[1]]$within.idx[[2]]))
+    dat$N_within <- length(dat$within_idx)
+    dat$both_idx <- array(Lp[[1]]$both.idx[[2]], length(Lp[[1]]$both.idx[[2]]))
+    dat$N_both <- length(dat$both_idx)
+    dat$N_lev <- c(length(dat$ov_idx1), length(dat$ov_idx2))
+    dat$between_idx <- c(dat$between_idx, sort(c(dat$within_idx, dat$both_idx)))
+
+    YLp <- lavobject@SampleStats@YLp # NB: one list entry per group
+    dat$mean_d <- do.call("rbind", lapply(YLp, function(x) do.call("rbind", x[[2]]$mean.d)))
+
+    cov_w <- array(unlist(lapply(YLp, function(x) x[[2]]$Sigma.W)), dim = c(ncol(dat$mean_d),
+                                                                            ncol(dat$mean_d),
+                                                                            Ng))
+    dat$cov_w <- aperm(cov_w, c(3, 1, 2))
+
+    cov_d <- unlist(lapply(YLp, function(x) x[[2]]$cov.d), recursive = FALSE)
+    for (i in 1:length(cov_d)) {
+      if (!inherits(cov_d[[i]], "matrix")) cov_d[[i]] <- with(dat,
+                                                              matrix(0, N_between + N_both + N_within,
+                                                                     N_between + N_both + N_within))
+    }
+    dat$cov_d <- cov_d
+    ## this evenly distributes loglik.x across unique cluster sizes; sums to correct value
+    llx <- sapply(YLp, function(x) x[[2]]$loglik.x)
+    dat$log_lik_x <- rep(llx / dat$ncluster_sizes, dat$ncluster_sizes)
+
+    ## clusterwise data summaries, for loo and waic and etc
+    cidx <- lavInspect(lavobject, 'cluster.idx')
+    if (inherits(cidx, "list")) {
+      if (length(cidx) > 1) {
+        for (g in 2:length(cidx)) {
+          cidx[[g]] <- cidx[[g]] + max(cidx[[(g - 1)]])
+        }
+      }
+      cidx <- unlist(cidx)
+    }
+    mean_d_full <- rowsum.default(as.matrix(dat$YX), cidx) / dat$cluster_size
+    if (dat$N_within > 0) {
+      for (i in 1:dat$N_within) {
+        mean_d_full[, dat$within_idx[i]] <- mean(as.matrix(dat$YX)[, dat$within_idx[i]])
+      }
+    }
+    tmpYX <- split.data.frame(dat$YX, cidx)
+    dat$YX <- do.call("rbind", tmpYX)
+    dat$log_lik_x_full <- llx_2l(Lp[[1]], dat$YX, mean_d_full, cidx)
+    dat$mean_d_full <- lapply(1:nrow(mean_d_full), function(i) mean_d_full[i, dat$between_idx])
+
+    ## cov_d is variability across clusters of same size, so irrelevant for clusterwise
+    ## (just send 0s to satisfy the Stan function)
+    cov_d_full <- lapply(1:length(dat$cluster_size), matrix, data = 0, nrow = NCOL(mean_d_full), ncol = NCOL(mean_d_full))
+    dat$cov_d_full <- cov_d_full
+
+    ## matrices for "saturated" logl
+    dat$xbar_w <- do.call("rbind", lapply(YLp, function(x) x[[2]]$Mu.W))
+    dat$xbar_b <- do.call("rbind", lapply(YLp, function(x) x[[2]]$Mu.B))
+    cov_b <- array(unlist(lapply(YLp, function(x) x[[2]]$Sigma.B)), dim = c(ncol(dat$xbar_b),
+                                                                           ncol(dat$xbar_b),
+                                                                           Ng))
+    dat$cov_b <- aperm(cov_b, c(3, 1, 2))
+    
+    ## cinv for each group, for computing saturated _rep matrices for ppp
+    dat$gs <- array(unlist(sapply(YLp, function(x) x[[2]]$s, simplify = FALSE)))
+  } else {
+    dat$nclus <- array(1, c(Ng, 2))
+    dat$cluster_size <- array(1, Ng)
+    dat$ncluster_sizes <- array(1, Ng)
+    dat$cluster_sizes <- array(1, Ng)
+    dat$cluster_size_ns <- array(1, Ng)
+    dat$between_idx <- array(0, 0)
+    dat$N_between <- 0
+    dat$within_idx <- array(0, 0)
+    dat$N_within <- 0
+    dat$both_idx <- array(0, 0)
+    dat$N_both <- 0
+    dat$ov_idx1 <- array(0, 0)
+    dat$ov_idx2 <- array(0, 0)
+    dat$p_tilde <- 0
+    dat$N_lev <- array(0, 2)
+    
+    dat$mean_d <- array(0, c(Ng, 0))
+    dat$cov_w <- array(0, c(Ng, 0, 0))
+    dat$log_lik_x <- array(0, Ng)
+    dat$log_lik_x_full <- array(0, Ng)
+    dat$cov_d <- array(0, c(Ng, 0, 0))
+    dat$mean_d_full <- array(0, c(Ng, 0))
+    dat$cov_d_full <- array(0, c(Ng, 0, 0))
+    dat$xbar_w <- array(0, c(Ng, 0))
+    dat$xbar_b <- array(0, c(Ng, 0))
+    dat$cov_b <- array(0, c(Ng, 0, 0))
+    dat$gs <- array(1, Ng)
+  } # multilevel
+  
   if (ord) {
     pta <- lav_partable_attributes(parTable(lavobject))
     ordidx <- pta$vidx$ov.ord[[1]]
@@ -1023,6 +1273,9 @@ lav2standata <- function(lavobject) {
     dat$Nordobs <- array(0, dat$Np)
     dat$OrdObsvar <- matrix(0, dat$Np, 0)
     dat$contidx <- array(1:nvar, nvar)
+    if (multilevel) {
+      dat$contidx <- array(1:ptot, ptot)
+    }
     dat$nlevs <- array(0, 0)
     dat$neach <- matrix(0, 0, 0)
   }
